/* Heuristica do vizinho mais proximo.
 * Este programa encontra uma solucao para o problema do caixeiro viajante utilizando o metodo do vizinho mais proximo.
 * O metodo e feito da seguinte forma: o caixeiro inicia em uma cidade qualquer e entao segue para a cidade mais proxima;
 * a partir desta, busca-se a cidade mais proxima, ainda nao visitada, e assim sucessivamente, ate que todas as cidades
 * sejam visitadas, retornando entao a cidade de origem.
 * Autora: Larissa Alves Petri     #USP: 5117594
 * e-mail: lariss@gmail.com
 * Data: /06/2006
 */


// Bibliotecas utilizadas:

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>


// Definicao das constantes:

#define N 6 // Numero de cidades do problema.
#define T 30 // Tamanho do nome do arquivo de entrada.
#define PI 3.141592 // Valor de pi.
#define R 6378.388 
#define INF 32000000 // Infinito.


// Declaracao das funcoes:

void le_arq (float x[N], float y[N], char c[N])
/* Funcao que le o arquivo de entrada, que  contem as coordenadas das cidades. Esse arquivo deve estar no formata txt.
 * x[N] e y[N]: vetores que receberao as coordenadas.
 * c[N]: nome do arquivo.
 */
{
	int i;
	FILE *e; // Ponteiro para o arquivo a ser aberto.
	
	// Testa se o arquivo foi aberto corretamente.
	if ((e = fopen (c, "r")) == NULL)
	{
		printf ("O arquivo %s nao pode ser aberto!\n", c); // Mensagem de erro!
		exit (0); // Sai da funcao.
	}

	// Le (do arquivo) as coordenadas e guarda-as em dois vetores, x e y.
	for (i = 0; i < N; i++)
		fscanf (e, "%*d %f %f\n", &x[i], &y[i]);

	// Fecha o arquivo.
	fclose (e);
}

void le_arq_matriz (int **d, char c[N])
/* Funcao que le o arquivo de entrada, que  contem as distancias entra as cidades. Esse arquivo deve ser .txt.
 * **d: matriz que recebera as distancias.
 * c[N]: nome do arquivo.
 */
{
	int i, j;
	FILE *e; // Ponteiro para o arquivo a ser aberto.
	
	// Testa se o arquivo foi aberto corretamente.
	if ((e = fopen (c, "r")) == NULL)
	{
		printf ("O arquivo %s nao pode ser aberto!\n", c); // Mensagem de erro!
		exit (0); // Sai da funcao.
	}

	// Le (do arquivo) as distancias e guarda-as na matriz d.
	for (i = 0; i < N; i++)
	{
		for (j = 0; j < (N - 1); j++)
			fscanf (e, " %d", &d[i][j]);
		fscanf (e, " %d\n", &d[i][j]);
	}

	// Fecha o arquivo.
	fclose (e);
}

void att (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia pseudo-euclidiana entre as cidades e guarda-a na matriz d (arquivos do tipo ATT).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j, t[N][N];
	float xd, yd, r[N][N];

	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			xd = x[i] - x[j];
			yd = y[i] - y[j];
			r[i][j] = sqrt ((xd * xd + yd * yd) / 10.0);
			t[i][j] = (int)(r[i][j]);
			if (t[i][j] < r[i][j])
				d[i][j] = t[i][j] + 1;
			else
				d[i][j] = t[i][j];
		}
}

void euc (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia euclidiana entre as cidades e guarda-a na matriz d (arquivos do tipo EUC_2D).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j;
	float xd, yd;
	
	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			xd = x[i] - x[j];
			yd = y[i] - y[j];
			d[i][j] = (int)(sqrt (xd * xd + yd * yd));
		}
}

void geo (float x[N], float y[N], int **d)
/* Funcao que calcula a distancia geometrica entre as cidades e guarda-a na matriz d (arquivos do tipo GEO).
 * x[N] e y[N]: vetores com as coordenadas.
 * **d: matriz que recebera o valor da distancia entre as cidades.
 */
{
	int i, j, deg;
	float min, q1, q2, q3;

	for (i = 0; i < N; i++)
	{
		deg = (int)(x[i]);
		min = x[i] - deg;
		x[i] = PI * (deg + 5.0 * min / 3.0) / 180.0;
		deg = (int)(y[i]);
		min = y[i] - deg;
		y[i] = PI * (deg + 5.0 * min / 3.0) / 180.0;
	}

	for (i = 0; i < N; i++)
		for (j = 0; j < N; j++)
		{
			q1 = cos (y[i] - y[j]);
			q2 = cos (x[i] - x[j]);
			q3 = cos (x[i] + x[j]);
			d[i][j] = (int)(R * acos (0.5 * ((1.0 + q1) * q2 - (1.0 - q1) * q3)) + 1.0);
		}
}

int verifica (int v[N], int e)
/* Funcao que verifica se o elemento a esta no vetor v.
 * v[N]: vetor solucao.
 * e: elemento do vetor.
 */
{
	int j = 0;

	// Percorre o vetor, procurando pelo elemento.
	while ((j < N) && (v[j] != e))
		j++;

	// Retorna 1 se o elemento esta no vetor e 0 caso contrario. 
	j = (j != N) ? 1 : 0;

	return j;
}

void insere_fim (int v[N], int e)
/* Funcao que insere o elemento e no final do vetor v.
 * v[N]: vetor solucao.
 * e: elemento que sera inserido no vetor.
 */
{
	int i = 0;

	// Percorre o vetor ate encontrar a primeira posicao vazia, ou seja, o primeiro elemento nulo.
	while ((v[i] != 0) && (i < N))
		i++;

	// Insere o elemento.
	v[i] = e;
}

int vmp (int v[N], int b, int **d)
/* Funcao que encontra o vertice mais proximo de b, que nao esteja no vetor v.
 * v[N]: vetor solucao inicial.
 * b: ultimo elemento, nao nulo, do vetor
 * **d: matriz de distancias.
 */
{
	int i, j, m;

	// Inicializacao da distancia.
	m = INF;

	// Percorre todas as cidades.
	for (i = 0; i < N; i++)
		// Verifica se a cidade em que esta nao e a cidade b.
		if (i != (b - 1))
			// Verifica se a cidade selecionada ja esta no vetor.
			if (!verifica (v, i + 1))
				// Verifica se esta cidade esta mais perto de b.
				if (m > d[b - 1][i])
				{
					// Atualiza a distancia.
					m = d[b - 1][i];
					// Atualiza a cidade mais proxima.
					j = i + 1;
				}

	// Retorna a cidade mais proxima.
	return j;
}

void sol_inicial (int v[N], int ind, int **d)
/* Funcao que encontra uma solucao inicial pelo metodo do vizinho mais proximo, comecando pelo vertice ind.
 * v[N]: vetor que recebera a solucao inicial.
 * ind: cidade da qual se iniciara a solucao.
 * **d: matriz de distancias.
 */
{
	int i, j, k; 
	
	// Inicializacao do vetor solucao.
	for (i = 0; i < N; i++)
		v[i] = 0;
	
	// A primeira posicao do vetor recebe a cidade inicial.
	v[0] = ind;
	// Encontra a cidade mais proxima da cidade inicial.
	k = vmp (v, ind, d);

	for (i = 1; i < N; i++)
	{
		// Insere a cidade mais proxima da ultima cidade inserida no vetor.
		insere_fim (v, k);
		// Encontra a cidade mais proxima da ultima cidade inserida no vetor.
		k = vmp (v, k, d);
	}
}

int main ()
{
	char c[T], tipo;
	int i, v[N], **d;
	float x[N], y[N];
	
	// Alocacao de memoria para a matriz de distancias.
	d = (int **) malloc (N * sizeof (int *));
	if (!d)
	{
		printf("Memoria Insuficiente!\n"); // Mensagem de erro.
		exit (0);
	}
	for(i=0; i<N; i++)
	{
		d[i] = (int *) malloc (N * sizeof (int));
		if (!d[i])
		{
			printf("Memoria Insuficiente!\n"); // Mensagem de erro.
			exit (0);
		}
	}
	
	// Leitura do arquivo de entrada e do tipo do arquivo.
	printf ("\n\nDigite o nome do arquivo de entrada: "); // <nome do arquivo>.txt
	scanf ("%s", c);
	fflush (stdin);
	printf ("\n\nInforme o tipo do arquivo (ATT = a, EUC = e, GEO = g, Matriz = m): ");
	scanf (" %c", &tipo);

	// Verificacao do tipo do arquivo e calculo da distancia.
	switch (tipo)
	{
		case 'a': le_arq (x, y, c);
			  att (x, y, d);
			  break;
		case 'e': le_arq (x, y, c);
			  euc (x, y, d);
			  break;
		case 'g': le_arq (x, y, c);
			  geo (x, y, d);
			  break;
		case 'm': le_arq_matriz (d, c);
	}

	// Obtencao da solucao inicial.
	sol_inicial (v, 1, d);

	// Imprime a solucao inicial.
	for (i = 0; i < N; i++)
		printf ("\nv[%d] = %d", i, v[i]);

	return 0;
}
